#![allow(clippy::missing_errors_doc)]

use std::{net::SocketAddr, time::Duration};

use bytes::{Bytes, BytesMut};
use lanspread_db::db::Game;
use lanspread_proto::{Message as _, Request, Response};
use lanspread_utils::maybe_addr;
use s2n_quic::{client::Connect, provider::limits::Limits, Client as QuicClient, Connection};
use tokio::sync::mpsc::{UnboundedReceiver, UnboundedSender};

static CERT_PEM: &str = include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/../../cert.pem"));

#[derive(Debug)]
pub enum ClientEvent {
    ListGames(Vec<Game>),
}

#[derive(Debug)]
pub enum ClientCommand {
    ListGames,
    GetGame(String),
    ServerAddr(SocketAddr),
}

async fn initial_server_alive_check(conn: &mut Connection) -> bool {
    let stream = match conn.open_bidirectional_stream().await {
        Ok(stream) => stream,
        Err(e) => {
            log::error!("failed to open stream: {e}");
            return false;
        }
    };

    let (mut rx, mut tx) = stream.split();

    // send ping
    if let Err(e) = tx.send(Request::Ping.encode()).await {
        log::error!("failed to send ping to server: {e}");
        return false;
    }
    let _ = tx.close().await;

    // receive pong
    if let Ok(Some(response)) = rx.receive().await {
        let response = Response::decode(response);
        if let Response::Pong = response {
            log::info!("server is alive");
            return true;
        }

        log::error!("server sent invalid response to ping: {response:?}");
    }

    false
}

#[allow(clippy::too_many_lines)]
pub async fn run(
    mut rx_control: UnboundedReceiver<ClientCommand>,
    tx_notify_ui: UnboundedSender<ClientEvent>,
) -> eyre::Result<()> {
    // blocking wait for remote address
    log::debug!("waiting for server address");
    let server_addr = loop {
        if let Some(ClientCommand::ServerAddr(addr)) = rx_control.recv().await {
            log::info!("got server address: {addr}");
            break addr;
        }
    };

    loop {
        let limits = Limits::default()
            .with_max_handshake_duration(Duration::from_secs(3))?
            .with_max_idle_timeout(Duration::from_secs(1))?;

        let client = QuicClient::builder()
            .with_tls(CERT_PEM)?
            .with_io("0.0.0.0:0")?
            .with_limits(limits)?
            .start()?;

        let conn = Connect::new(server_addr).with_server_name("localhost");
        let mut conn = match client.connect(conn).await {
            Ok(conn) => conn,
            Err(e) => {
                log::error!("failed to connect to server: {e}");
                tokio::time::sleep(Duration::from_secs(3)).await;
                continue;
            }
        };

        conn.keep_alive(true)?;

        if !initial_server_alive_check(&mut conn).await {
            continue;
        }

        log::info!(
            "connected: (server: {}) (client: {})",
            maybe_addr!(conn.remote_addr()),
            maybe_addr!(conn.local_addr())
        );

        // tx
        while let Some(cmd) = rx_control.recv().await {
            let request = match cmd {
                ClientCommand::ListGames => Request::ListGames,
                ClientCommand::GetGame(id) => Request::GetGame { id },
                ClientCommand::ServerAddr(_) => Request::Invalid(
                    Bytes::new(),
                    "invalid control message (ServerAddr), should not happen".into(),
                ),
            };

            let data = request.encode();
            log::debug!("encoded data: {}", String::from_utf8_lossy(&data));

            let stream = match conn.open_bidirectional_stream().await {
                Ok(stream) => stream,
                Err(e) => {
                    log::error!("failed to open stream: {e}");
                    break;
                }
            };

            let (mut rx, mut tx) = stream.split();

            if let Err(e) = tx.send(data).await {
                log::error!("failed to send request to server {:?}", e);
            }

            let mut data = BytesMut::new();
            while let Ok(Some(bytes)) = rx.receive().await {
                data.extend_from_slice(&bytes);
            }
            log::debug!("{} bytes received from server", data.len());
            log::trace!("msg: (raw): {}", String::from_utf8_lossy(&data));

            let response = Response::decode(data.freeze());
            log::trace!("msg: {response:?}");

            match response {
                Response::Games(games) => {
                    for game in &games {
                        log::trace!("{game}");
                    }

                    if let Err(e) = tx_notify_ui.send(ClientEvent::ListGames(games)) {
                        log::debug!("failed to send ClientEvent::ListGames to client {e:?}");
                    } else {
                        log::info!("sent ClientEvent::ListGames to Tauri client");
                    }
                }
                Response::Game(game) => log::debug!("game received: {game:?}"),
                Response::GameNotFound(id) => log::debug!("game not found {id}"),
                Response::InvalidRequest(request_bytes, err) => log::error!(
                    "server says our request was invalid (error: {}): {}",
                    err,
                    String::from_utf8_lossy(&request_bytes)
                ),
                Response::EncodingError(err) => {
                    log::error!("server encoding error: {err}");
                }
                Response::DecodingError(data, err) => {
                    log::error!(
                        "response decoding error: {} (data: {})",
                        err,
                        String::from_utf8_lossy(&data)
                    );
                }
                Response::Pong => (), // ignore (should never happen)
            }

            if let Err(err) = tx.close().await {
                log::error!("failed to close stream: {err}");
            }
        }
    }
}
