use std::{
    net::SocketAddr,
    path::{Path, PathBuf},
    sync::Arc,
};

use lanspread_client::{ClientCommand, ClientEvent};
use lanspread_db::db::{Game, GameDB};
use lanspread_mdns::{discover_service, LANSPREAD_INSTANCE_NAME, LANSPREAD_SERVICE_TYPE};
use tauri::{AppHandle, Emitter as _, Manager};
use tokio::sync::{mpsc::UnboundedSender, Mutex};

// Learn more about Tauri commands at https://tauri.app/develop/calling-rust/

struct LanSpreadState {
    server_addr: Mutex<Option<SocketAddr>>,
    client_ctrl: UnboundedSender<ClientCommand>,
    games: Arc<Mutex<GameDB>>,
}

#[tauri::command]
fn request_games(state: tauri::State<LanSpreadState>) {
    log::debug!("request_games");

    if let Err(e) = state.inner().client_ctrl.send(ClientCommand::ListGames) {
        log::error!("Failed to send message to client: {e:?}");
    }
}

#[tauri::command]
fn run_game_backend(id: String, state: tauri::State<LanSpreadState>) -> String {
    log::error!("Running game with id {id}");

    // let result = Command::new(r#"C:\Users\ddidderr\scoop\apps\mpv\0.39.0\mpv.exe"#).spawn();

    if let Err(e) = state.inner().client_ctrl.send(ClientCommand::GetGame(id)) {
        log::error!("Failed to send message to client: {e:?}");
    }

    "TODO".to_string()

    // if result.is_ok() {
    // "Ok".to_string()
    // } else {
    // "Failed to run game".to_string()
    // }
}

fn set_game_install_state_from_path(game_db: &mut GameDB, path: &Path, installed: bool) {
    if let Some(file_name) = path.file_name() {
        if let Some(file_name) = file_name.to_str() {
            if let Some(game) = game_db.get_mut_game_by_id(file_name) {
                if installed {
                    log::error!("Game is installed: {game}");
                } else {
                    log::error!("Game is missing: {game}");
                }
                game.installed = installed;
            }
        }
    }
}

#[tauri::command]
fn update_game_directory(app_handle: tauri::AppHandle, path: String) {
    let path = PathBuf::from(path);
    if !path.exists() {
        log::error!("game dir {path:?} does not exist");
    }

    let entries = match path.read_dir() {
        Ok(entries) => entries,
        Err(e) => {
            log::error!("Failed to read game dir: {e}");
            return;
        }
    };

    tauri::async_runtime::spawn(async move {
        let mut game_db = app_handle
            .state::<LanSpreadState>()
            .inner()
            .games
            .lock()
            .await;

        // Reset all games to uninstalled
        game_db.set_all_uninstalled();

        // update game_db with installed games from real game directory
        entries.into_iter().for_each(|entry| {
            if let Ok(entry) = entry {
                if let Ok(path_type) = entry.file_type() {
                    if path_type.is_dir() {
                        let path = entry.path();
                        if path.join(".softlan_game_installed").exists() {
                            set_game_install_state_from_path(&mut game_db, &path, true);
                        }
                    }
                }
            }
        });

        if let Err(e) = app_handle.emit("games-list-updated", Some(game_db.all_games())) {
            log::error!("Failed to emit games-list-updated event: {e}");
        }
    });
}

async fn find_server(app: AppHandle) {
    log::info!("Looking for server...");

    loop {
        match discover_service(LANSPREAD_SERVICE_TYPE, Some(LANSPREAD_INSTANCE_NAME)) {
            Ok(server_addr) => {
                log::info!("Found server at {server_addr}");
                let state: tauri::State<LanSpreadState> = app.state();
                {
                    // mutex scope
                    let mut addr = state.server_addr.lock().await;
                    *addr = Some(server_addr);
                }
                state
                    .client_ctrl
                    .send(ClientCommand::ServerAddr(server_addr))
                    .unwrap();
                request_games(state);
                break;
            }
            Err(e) => {
                log::warn!("Failed to find server: {e} - retrying...");
            }
        }
    }
}

async fn update_game_db(games: Vec<Game>, app: AppHandle) {
    for game in &games {
        log::trace!("client event ListGames iter: {game:?}");
    }

    let state = app.state::<LanSpreadState>();

    // Store games list
    let mut state_games = state.games.lock().await;
    *state_games = GameDB::from(games.clone());

    // Tell Frontend about new games list
    if let Err(e) = app.emit("games-list-updated", Some(games)) {
        log::error!("Failed to emit games-list-updated event: {e}");
    } else {
        log::info!("Emitted games-list-updated event");
    }
}

#[cfg_attr(mobile, tauri::mobile_entry_point)]
pub fn run() {
    let tauri_logger_builder = tauri_plugin_log::Builder::new()
        .clear_targets()
        .target(tauri_plugin_log::Target::new(
            tauri_plugin_log::TargetKind::Stdout,
        ))
        .level(log::LevelFilter::Info)
        .level_for("lanspread_client", log::LevelFilter::Debug)
        .level_for("lanspread_tauri_leptos_lib", log::LevelFilter::Debug)
        .level_for("mdns_sd::service_daemon", log::LevelFilter::Off);

    // channel to pass commands to the client
    let (tx_client_control, rx_client_control) =
        tokio::sync::mpsc::unbounded_channel::<ClientCommand>();

    // channel to receive events from the client
    let (tx_client_event, mut rx_client_event) =
        tokio::sync::mpsc::unbounded_channel::<ClientEvent>();

    let lanspread_state = LanSpreadState {
        server_addr: Mutex::new(None),
        client_ctrl: tx_client_control,
        games: Arc::new(Mutex::new(GameDB::empty())),
    };

    tauri::Builder::default()
        .plugin(tauri_plugin_store::Builder::new().build())
        .plugin(tauri_plugin_dialog::init())
        .plugin(tauri_logger_builder.build())
        .plugin(tauri_plugin_shell::init())
        .invoke_handler(tauri::generate_handler![
            run_game_backend,
            request_games,
            update_game_directory
        ])
        .manage(lanspread_state)
        .setup(|app| {
            let app_handle = app.handle().clone();
            // discover server
            tauri::async_runtime::spawn(async move { find_server(app_handle).await });
            tauri::async_runtime::spawn(async move {
                lanspread_client::run(rx_client_control, tx_client_event).await
            });

            let app_handle = app.handle().clone();
            tauri::async_runtime::spawn(async move {
                while let Some(event) = rx_client_event.recv().await {
                    match event {
                        ClientEvent::ListGames(games) => {
                            log::debug!("Received client event: ListGames");
                            update_game_db(games, app_handle.clone()).await;
                        }
                    }
                }
            });

            Ok(())
        })
        .run(tauri::generate_context!())
        .expect("error while running tauri application");
}
